from typing import Optional

from transformers import BertTokenizer

from transformersx.data import *


class TestDataProcessor(TaskDataProcessor):
    def __generate_examples(self, count, type):
        return [TaskInputExample(guid=str(i), text_a=type + '_text_a_' + str(i), text_b=type + '_text_b_' + str(i),
                                 label=str(i % 2)) for i in range(count)]

    def get_train_examples(self, limit_length: Optional[int] = None):
        return self.__generate_examples(30, 'train')

    def get_eval_examples(self, limit_length: Optional[int] = None):
        return self.__generate_examples(10, 'eval')

    def get_test_examples(self, limit_length: Optional[int] = None):
        return self.__generate_examples(5, 'test')


class TestDatasetFactory:
    def __int__(self):
        data_args = TaskDataArguments()
        data_args.data_dir = '.'
        self.data_store = LocalDataStore("test", data_args)
        self.data_convertor = TaskDataConverter(BertTokenizer('../../dataset/vocab.txt'), ['0', '1'])
        self.dataset_factory = TaskDatasetFactory(self.data_store, TestDataProcessor(), self.data_convertor)

    def test_create_dataset(self):
        train_dataset = self.dataset_factory.create_train_dataset()
        eval_dataset = self.dataset_factory.create_eval_dataset()
        test_dataset = self.dataset_factory.create_predict_dataset()

    def test_load_dataset(self):
        pass



from box import Box
from sanic import Sanic
from sanic_openapi import swagger_blueprint, doc
from .service_args import ServiceArgs
from sanic.response import json, text
from dataclasses import dataclass, asdict, fields
from ai_transformersx import log, get_transformers_model

types = dict([
    (int, doc.Integer),
    (float, doc.Float),
    (str, doc.String),
    (bool, doc.Boolean)
])


def to_dict(dataclass):
    json_body = {}
    for f in fields(dataclass):
        json_body[f.name] = types.get(f.type)()
    return json_body


class TaskService:
    def __init__(self, args: ServiceArgs = ServiceArgs()):
        self._args = args
        self._model = None
        self.__init()

    def __init(self):
        self.app = Sanic()
        self.app.blueprint(swagger_blueprint)

    def run_task_service(self):
        log.info("Start Task Service with arguments: {}".format(str(self._args)))

        self.app.run(self._args.host, self._args.port)

    def __get_model(self):
        if not self._model:
            self._model = get_transformers_model(self._args.models_dir, self._args.model_name)
        return self._model

    def predict(self, *input):
        return self.__get_model().predict(*input)

    def task(self, uri, requestData, responseData, summary="Task"):
        post = self.app.post(uri)

        def response(handler):
            async def handle_request(request):
                # log.info("Accept request: {}".format(request.json))
                return json(asdict(handler(Box(request.json))))

            _, newHandler = post(handle_request)
            doc.summary(summary)
            doc.consumes(doc.JsonBody(to_dict(requestData)), content_type="application/json", location='body')(
                newHandler)
            doc.produces(to_dict(responseData), content_type="application/json")(newHandler)
            return newHandler

        return response



from typing import Optional


class TaskDataProcessor:
    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, limit_length: Optional[int] = None):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_eval_examples(self, limit_length: Optional[int] = None):
        """Gets a collection of `InputExample`s for the eval set."""
        raise NotImplementedError()

    def get_test_examples(self, limit_length: Optional[int] = None):
        """Gets a collection of `InputExample`s for the test set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()



from collections import OrderedDict

from transformers import is_tf_available
from transformers.modeling_roberta import RobertaConfig

from transformers.tokenization_roberta import RobertaTokenizer, RobertaTokenizerFast

from ..model_base import TaskModels, ModelType, ModelTaskType, model_func

default_model = model_func(ModelType.roberta, RobertaConfig, RobertaTokenizer, 'roberta')
fast_model = model_func(ModelType.roberta, RobertaConfig, RobertaTokenizerFast, 'roberta')


class Roberta_Task_Models_Base(TaskModels):
    MODELS = {
        "cn": [default_model("clue/roberta_chinese_clue_tiny"),
               default_model("clue/roberta_chinese_3L312_clue_tiny"),
               default_model("clue/roberta_chinese_pair_tiny"),
               default_model("clue/roberta_chinese_clue_base"),
               default_model("clue/roberta_chinese_large"),
               default_model("clue/roberta_chinese_clue_large"),
               default_model("clue/roberta_chinese_pair_large"),
               default_model("roberta_chinese_3L768_clue_tiny"),
               default_model("chinese-roberta-wwm-ext-large"),
               default_model("onePatient/roberta_chinese_clue_tiny"),
               default_model("hfl/chinese-roberta-wwm-ext"),
               ]}


class Roberta_Task_Models(Roberta_Task_Models_Base):
    from transformers.modeling_roberta import RobertaModel, RobertaForMaskedLM, RobertaForMultipleChoice, \
        RobertaForQuestionAnswering, RobertaForSequenceClassification, RobertaForTokenClassification
    MODEL_CLASSES = OrderedDict([
        (ModelTaskType.base, RobertaModel),
        (ModelTaskType.lm_head, RobertaForMaskedLM),
        (ModelTaskType.seq_cls, RobertaForSequenceClassification),
        (ModelTaskType.token_cls, RobertaForTokenClassification),
        (ModelTaskType.qa, RobertaForQuestionAnswering),
        (ModelTaskType.multi_choice, RobertaForMultipleChoice)
    ])


if is_tf_available():
    class TFRoberta_Task_Models(Roberta_Task_Models_Base):
        from transformers.modeling_tf_roberta import TFRobertaModel, TFRobertaForMaskedLM, \
            TFRobertaForQuestionAnswering, TFRobertaForSequenceClassification, TFRobertaForTokenClassification
        MODEL_CLASSES = OrderedDict([
            (ModelTaskType.base, TFRobertaModel),
            (ModelTaskType.lm_head, TFRobertaForMaskedLM),
            (ModelTaskType.seq_cls, TFRobertaForSequenceClassification),
            (ModelTaskType.token_cls, TFRobertaForTokenClassification),
            (ModelTaskType.qa, TFRobertaForQuestionAnswering),
        ])



from .tasks import *

from .examples_management import ExampleManagement

task_manager = ExampleManagement()
task_manager.register_tasks([
    ('news', NewsSegmentTask),
    ('news_cls', NewsClsTask),
    ('sentiment', SentimentTask)
])

if is_turbo_available():
    task_manager.register_task('turbo-news', TurboNewsSegmentTask)

if __name__ == "__main__":
    task_manager.start_example_task()



from typing import Dict, NamedTuple, Optional, List, Union
import numpy as np
import torch
from dataclasses import dataclass
from transformers import DataCollator
from transformers.data.data_collator import InputDataClass


class TaskEvalPrediction(NamedTuple):
    predictions: np.ndarray
    label_ids: np.ndarray


class TaskPredictionOutput(NamedTuple):
    guids: np.ndarray
    predictions: np.ndarray
    label_ids: Optional[np.ndarray]
    eval_losses: np.ndarray
    metrics: Optional[Dict[str, float]] = None


class TaskTrainOutput(NamedTuple):
    global_step: int
    training_loss: float


@dataclass
class TaskDefaultDataCollatorx(DataCollator):
    def collate_batch(self, features: List[InputDataClass]) -> Dict[str, torch.Tensor]:
        first = features[0]

        # Special handling for labels.
        # Ensure that tensor is created with the correct type
        # (it should be automatically the case, but let's make sure of it.)
        if hasattr(first, "label") and first.label is not None:
            if type(first.label) is int:
                labels = torch.tensor([f.label for f in features], dtype=torch.long)
            else:
                labels = torch.tensor([f.label for f in features], dtype=torch.float)
            batch = {"labels": labels}
        elif hasattr(first, "label_ids") and first.label_ids is not None:
            if type(first.label_ids[0]) is int:
                labels = torch.tensor([f.label_ids for f in features], dtype=torch.long)
            else:
                labels = torch.tensor([f.label_ids for f in features], dtype=torch.float)
            batch = {"labels": labels}
        elif hasattr(first, 'guid') and first.guid is not None:
            guids = torch.tensor([f.guid for f in features], dtype=torch.long)
            batch = {"guid": guids}
        else:
            batch = {}

        # Handling of all other possible attributes.
        # Again, we will use the first element to figure out which key/values are not None for this model.
        for k, v in vars(first).items():
            if k not in ("guid", "label", "label_ids") and v is not None and not isinstance(v, str):
                batch[k] = torch.tensor([getattr(f, k) for f in features], dtype=torch.long)
        return batch


PREFIX_CHECKPOINT_DIR = "checkpoint"



import torch
from transformers import PreTrainedModel, get_linear_schedule_with_warmup, AdamW

from transformersx.train.training_args import TaskTrainingArguments
from .trainer_base import TrainerBase


class TaskTrainerOptimizers(TrainerBase):
    def __init__(self, args: TaskTrainingArguments,
                 model: PreTrainedModel,
                 optimizer: torch.optim.Optimizer = None,
                 scheduler: torch.optim.lr_scheduler.LambdaLR = None):
        super().__init__(args)
        self._model = model
        self._optimizer = optimizer
        self._scheduler = scheduler

    def get_model(self) -> PreTrainedModel:
        return self._model

    def get_optimizer(self) -> torch.optim.Optimizer:
        if not self._optimizer:
            self._optimizer = self._create_optimizer()
        return self._optimizer

    def get_scheduler(self, num_training_steps=-1) -> torch.optim.lr_scheduler.LambdaLR:
        if not self._scheduler:
            self._scheduler = self._create_scheduler(self.get_optimizer(), num_training_steps)

        return self._scheduler

    def set_for_start_train(self):
        self._model.to(self._args.device)
        if self._args.fp16:
            if not self.is_apex_available():
                raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
            self._model, self._optimizer = self.amp.initialize(self._model, self._optimizer,
                                                               opt_level=self._args.fp16_opt_level)

            # multi-gpu training (should be after apex fp16 initialization)
        if self._args.n_gpu > 1:
            self._model = torch.nn.DataParallel(self._model)

        # Distributed training (should be after apex fp16 initialization)
        if self._args.local_rank != -1:
            self._model = torch.nn.parallel.DistributedDataParallel(
                self._model,
                device_ids=[self._args.local_rank],
                output_device=self._args.local_rank,
                find_unused_parameters=True,
            )
        self._model.zero_grad()

    def _create_optimizer(self):
        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self._model.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": self._args.weight_decay,
            },
            {
                "params": [p for n, p in self._model.named_parameters() if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]
        return AdamW(optimizer_grouped_parameters, lr=self._args.learning_rate, eps=self._args.adam_epsilon)

    def _create_scheduler(self, optimizer, num_training_steps):
        return get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=self._args.warmup_steps, num_training_steps=num_training_steps
        )



from ..model import ALL_TASK_MODEL_PATHS
from ..transformersx_base import *


@configclass
class FastTaskArguments:
    action: str = field("train", "the task action: train,eval,predict")
    data_dir: str = field("/app/dataset", "the data directory for train,eval,test,label")
    model_finetuned_dir: str = field("/app/models/finetuning",
                                     "The 'model_base_dir' generally is for the finetuned models."
                                     "Generally, the model is loaded from 'model_finetuned_dir' firstly."
                                     " If the model cannot be found, it will be loaded from the model_pretrained_dir.")

    model_pretrained_dir: str = field("/app/models/pretrained",
                                      "This folder is for the pretrained models downloaded from Internet.")
    model_name: str = field("bert-base-chinese", "the name of model: " + str(ALL_TASK_MODEL_PATHS))
    train_file: str = field('train.csv', "the train data file name")
    eval_file: str = field('dev.csv', "the eval data file name")
    labels_file: str = field('labels.csv', "the test data file name")
    text_col: str = field('text', '')
    label_col: str = field('label', '')
    batch_size_per_gpu: int = field(16, '')
    learning_rate: float = field(5e-5, "The initial learning rate for Adam.")
    num_train_epochs: int = field(3, "Total number of training epochs to perform.")
    max_seq_length: int = field(512, '')
    multi_label: bool = field(False, '')
    model_type: str = field('bert', '')
    fp16: bool = field(False, '')



from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    metadata_version='2.1',
    name='ai-transformersx',
    version='0.5.28',
    description='tools for using huggingface/transformers more easily',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/aicanhelp/ai-transformers',
    author='modongsong',
    author_email='modongsongml@163.com',
    license='MIT',

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='deeplearning tools development',  # Optionalz
    packages=find_packages(),  # Required
    python_requires='>=3.5, <4',
    install_requires=[],  # Optional

    extras_require={  # Optional
        'dev': ['check-manifest'],
        'test': ['coverage', 'pytest'],
    },

    # package_data={  # Optional
    #     'sample': ['package_data.dat'],
    # },

    # data_files=[('my_data', ['data/data_file'])],  # Optional

    # entry_points={  # Optional
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },

    project_urls={  # Optional
        'Bug Reports': 'https://github.com/aicanhelp/ai-transformersx/issues',
        'Source': 'https://github.com/aicanhelp/ai-transformers/',
    },
)



from ai_transformersx.service.task_service import TaskService
from dataclasses import dataclass, asdict, field, fields

service = TaskService()


@dataclass
class RequestData:
    content: str = field(default="content")


@dataclass
class ResponseData:
    content: str = field(default="test")


@service.task('/test', RequestData, ResponseData)
def task_request(request_json):
    print(request_json)
    return ResponseData("test")


class TestDemoService():

    def test_request(self):
        request, response = service.app.test_client.post('/test', data='{"content":"content"}')
        assert request.json['content'] == 'content' and response.json['content'] == 'test'

    def test_run_service(self):
        service.run_task_service()

    def test_dataclass(self):
        for f in fields(RequestData):
            print(f.name, f.type == str)
